//Program to estimate a pairwise log likelihood surface for finite populations in a crossing experiment.

#include "utilities.h"
using namespace std;


void sample_distribution_ind(int Npop, int Ncrosses, double rho, vector < qSample > &mcSample, int Nmc);
double log_likehood_estimate(vector < qSample > &mcSample, vector <int> &draw);
void combined_resample_cross(vector < std::string > &population, vector < std::string > &populationNew, double rho);
void init_population(vector < std::string > &population, int N, int n_loci);

int main(int argc, char *argv[]){
	
	if(argc != 8){
		cout<< "Usage ./run_finite Ngrid rMax Ncrosses Npop Nmc Nsample seed" <<endl;
		exit(0);
	}
	
	int Ngrid=(int)atof(argv[1]); /* number of points in a rho grid */
	double rMax=atof(argv[2]); /* maximum rho considered */
	int Ncrosses=(int)atof(argv[3]); /* number of crossing generations */
	int Npop=(int)atof(argv[4]); /* population size */
	int Nmc=(int)atof(argv[5]); /* number of samples per rho value */
	int Nsample=(int)atof(argv[6]); /* number of haplotypes drawn from the final pool */
	int seed=(int)atof(argv[7]); /* seed for the  random number generators */
		
	gsl_rng_env_setup();
	srand((unsigned int)(seed)); 
	gsl_rng *rgen = gsl_rng_alloc (gsl_rng_taus);
	gsl_rng_set (rgen, seed);
	
	vector < double> rhoV;
	for(int i=0; i< Ngrid; i++){
		double rho=i*rMax/(double)(Ngrid-1);
		rhoV.push_back(rho);
	}
			
	vector < observation > table;
	create_pairs(Nsample, table);
	
	int N=0;
	for(int i=0; i< table[0].draw.size(); i++){
		N+=table[0].draw[i];
	}
	int Npairs=table.size();
	
	
	vector < vector< double > > llV(table.size(), vector<double>(rhoV.size()));
	vector < qSample > mcSample;
	
	for(int ri=0; ri<rhoV.size(); ri++){
		
		sample_distribution_ind(Npop,Ncrosses,rhoV[ri],mcSample,Nmc);
		for(int i=0; i< table.size(); i++){	
			double val=log_likehood_estimate(mcSample,table[i].draw);
			llV[i][ri]=val;
		}
	};
	
	/* printing done in the end */
	print_header(N,Npairs,Ngrid,rhoV);
	for(int i=0; i< table.size(); i++){	
		print_line(i+1, table[i].draw, llV[i]);
	}
	return EXIT_SUCCESS;
};



/* given the samples from finite population crossing estimate log likehood of an observation draw */
double log_likehood_estimate(vector < qSample > &mcSample, vector <int> &draw){
	
	double z=0.0;
	double minL=0.0;
	
	/* this is to avoid working with very small numbers */
	for(int i=0; i<mcSample.size(); i++){
		if(log_likehood(draw,mcSample[i].q)<minL){
			minL=log_likehood(draw,mcSample[i].q);
		}
	}		
	/* summing over the sample from the distribution*/
	for(int i=0; i<mcSample.size(); i++){
		z+=exp(log_likehood(draw,mcSample[i].q)-minL);
	}
	z=z/(double)mcSample.size();
	
	return (log(z)+minL);
		
}


void sample_distribution_ind(int Npop, int Ncrosses, double rho, vector < qSample > &mcSample, int Nmc){
	
	mcSample.resize(0);
	vector < int > Nv(4);
	
	double eps=0.001;
	
	
	/* generate Nmc samples from the finite size population crossing, these will be used to estimate the loglikelihood of observations */
	for(int mcI=0; mcI< Nmc; mcI++){
		
		vector < std::string > population;
		vector < std::string  > populationNew;
		
		init_population(population,Npop,2);
		populationNew=population;
		for(int iC=0; iC<Ncrosses; iC++){
			combined_resample_cross(population,populationNew,rho);
			population=populationNew;	
		}
	
		double qij[4];
		for(int i=0; i<4; i++){
			qij[i]=0.0;
		}
		
		for(int n=0; n< population.size(); n++){
			
			if(population[n] == "00"){
				qij[0]+=1.0;
			}
			
			if(population[n] == "01"){
				qij[1]+=1.0;
			}
			
			if(population[n] == "01"){
				qij[2]+=1.0;
			}
			
			if(population[n] == "11"){
				qij[3]+=1.0;
			}
						
		}
		for(int i=0; i<4; i++){
			qij[i]=qij[i]/(double)population.size();
		}
		
		/* for zero entries add epsilon and renormalize accordingly to avoid -infinities in the log likehood for zero recombination case */
		for(int i=0; i<4; i++){
			if(qij[i]<eps){
				qij[i]=eps;
			}
		}
		
		double z=0.0;
		for(int i=0; i< 4; i++){
			z+=qij[i];
		}
		for(int i=0; i< 4; i++){
			qij[i]=qij[i]/z;
		}
		
		qSample obs;
		for(int i=0; i<4; i++){
			obs.q.push_back(qij[i]);
		}
	
		mcSample.push_back(obs);
		
	}			
	
};



/* this needs to be modified if more than two founder strains or unequal numbers of two parental strains form the inital pool */
void init_population(vector < std::string > &population, int N, int n_loci){
		
	population.clear();
	
	std::string ones;
	std::string zeros;
	
	for(int l=0;l<n_loci; l++){
		ones.append("1");
		zeros.append("0");
	}
	
	for(int n=0;n<N/2; n++){
		population.push_back(ones);
		population.push_back(zeros);
	}
	
}


/* draw new generation with replacement and recombine with rate rho */
 
void combined_resample_cross(vector < std::string > &population, vector < std::string > &populationNew, double rho){
	
	int N=population.size();
		
	std::string recombinant1;
	std::string recombinant2;
	std::string recombinant1N;
	std::string recombinant2N;
	
	int k=0;
	for(int n=0; n< N/2; n++){
	
		/* two random parent haplotypes */
		double rn=(double)(rand())/(double)(RAND_MAX);
		int ind1=(int)(rn*(double)(N));
		
		rn=(double)(rand())/(double)(RAND_MAX);
		int ind2=(int)(rn*(double)(N));
		
		rn=(double)(rand())/(double)(RAND_MAX);
		
		recombinant1=population[ind1];
		recombinant2=population[ind2];
		
		recombinant1N=recombinant1;
		recombinant2N=recombinant2;
		
		/* check if they recombine or not */
		if(rn<rho){
			recombinant1N[1]=recombinant2[1];
			recombinant2N[1]=recombinant1[1];
		}
		populationNew[k++]=recombinant1N;
		populationNew[k++]=recombinant2N;
				
	}
	
}

							
					
 
							
	
